#!/bin/awk -f
#!/opt/busybox/awk -f

# EXPERIMENTAL Markdown processor with minimal dependencies.
# Meant to support all features of John Grubers basic Markdown
# + a number of common extensions, mostly inspired by Pandoc Markdown
#
# ToDo:
# - HTML processing / escaping (according to environment flag)
# - em-dashes and arrows

# Supported Features / TODO:
# ==========================
# [x] done    [ ] todo    [-] not planned    ? unsure
#
# Basic Markdown - Block elements:
# -------------------------------
# - [x] Paragraphs
#   - [x] Double space line breaks
# - [x] Proper block element nesting
# - [x] Headings
# - [x] ATX-Style Headings
# - [x] Blockquotes
# - [x] Lists (ordered, unordered)
# - [x] Code blocks (using indention)
# - [x] Horizontal rules
# - [x] Verbatim HTML block (disabled by default)
#
# Basic Markdown - Inline elements:
# ---------------------------------
# - [x] Links
# - [x] Reference style links
# - [x] Emphasis *em*/**strong** (*Asterisk*, _Underscore_)
# - [x] `code`, also ``code containing `backticks` ``
# - [x] Images / reference style images
# - [x] <automatic links>
# - [x] backslash escapes
# - [x] Verbatim HTML inline (disabled by default)
# - [x] HTML escaping
#
# NOTE: Set the environment variable MD_HTML=true to enable verbatim HTML
#
# Extensions - Block elements:
# ----------------------------
# -  ?  Heading identifiers (php md, pandoc)
# - [x] Automatic heading identifiers (custom)
# - [x] Fenced code blocks (php md, pandoc)
#   - [x] Fenced code attributes
# - [ ] Tables
#   -  ?  Simple table (pandoc)
#   -  ?  Multiline table (pandoc)
#   -  ?  Grid table (pandoc)
#   -  ?  Pipe table (php md pandoc)
# - [x] Line blocks (pandoc)
# - [x] Task lists (pandoc)
# - [ ] Definition lists (php md, pandoc)
# - [-] Numbered example lists (pandoc)
# - [-] Metadata blocks (pandoc)
# - [x] Fenced Divs (pandoc)
#
# Extensions - Inline elements:
# ----------------------------
# - [x] Ignore embedded_underscores (php md, pandoc)
# - [x] ~~strikeout~~ (pandoc)
# - [x] ^Superscript^ ~Subscript~ (pandoc)
# - [-] Bracketed spans (pandoc)
#   - [-] Inline attributes (pandoc)
# - [-] TEX-Math (pandoc)
# -  ?  Footnotes (php md)
# -  ?  Abbreviations (php md)
# -  ?  "Curly quotes" (smartypants)
# - [ ] em-dashes (--) (smartypants old)
# -  ?  ... three-dot ellipsis (smartypants)
# - [-] en-dash (smartypants)
# - [ ] Automatic em-dash / en-dash
# - [ ] Automatic -> Arrows <-

function debug(text) { printf "\n---\n%s\n---\n", text > "/dev/stderr"; }

function HTML ( text ) {
  gsub( /&/,  "\\&amp;",  text );
  gsub( /</,  "\\&lt;",   text );
  gsub( />/,  "\\&gt;",   text );
  gsub( /"/,  "\\&quot;", text );
  gsub( /'/,  "\\&#x27;", text );
  gsub( /\\/, "\\&#x5C;", text );
  return text;
}

function inline( line, LOCAL, len, code, href, guard ) {
  nu = "(\\\\\\\\|\\\\[^\\\\]|[^\\\\_]|_[[:alnum:]])*"    # not underline (except when escaped)
  na = "(\\\\\\\\|\\\\[^\\\\]|[^\\\\\\*])*"  # not asterisk (except when escaped)
  ieu =  "_([^_[:space:]]|[^_[:space:]]" nu "[^_[:space:]])_"                 # inner <em> (underline)
  isu = "__([^_[:space:]]|[^_[:space:]]" nu "[^_[:space:]])__"                # inner <strong> (underline)
  iea =    "\\*([^\\*[:space:]]|[^\\*[:space:]]" na "[^\\*[:space:]])\\*"     # inner <em> (asterisk)
  isa = "\\*\\*([^\\*[:space:]]|[^\\*[:space:]]" na "[^\\*[:space:]])\\*\\*"  # inner <strong> (asterisk)

  if ( line ~ /^$/ ) {  # Recursion End
    return "";

  #  omit processing of escaped characters
  } else if ( line ~ /^\\[]\\`\*_\{\}\(\)#\+-\.![]/) {
    return substr(line, 2, 1) inline( substr(line, 3) );

  # hard brakes
  } else if ( match(line, /^  \n/) ) {
    return "<br />\n" inline( substr(line, RLENGTH + 1) );

  #  ``code spans``
  } else if ( match( line, /^`+/) ) {
    len = RLENGTH
    guard = substr( line, 1, len )
    if ( match(line, guard ".*" guard) ) {
      code = substr( line, len + 1, match( substr(line, len + 1), guard ) - 1)
      len = 2 * length(guard) + length(code)
      #  strip single surrounding white spaces
      code = gensub( / (.*) /, "\\1", "1" , code)
      #  escape HTML within code span
      gsub( /&/, "\\&amp;", code ); gsub( /</, "\\&lt;", code ); gsub( />/, "\\&gt;", code );
      return "<code>" code "</code>" inline( substr( line, len + 1 ) )
    }

  #  quick links ("automatic links" in md doc)
  } else if ( match( line, /^<[a-zA-Z]+:\/\/([-\.[:alnum:]]+)(:[0-9]*)?(\/[^>]*)?>/ ) ) {
    len = RLENGTH;
    href = HTML( substr( line, 2, len - 2) );
    return "<a href=\"" href "\">" href "</a>" inline( substr( line, len + 1) );

  # inline links
  } else if ( match(line, /^\[([^]]+)\]\(([^"\)]+)([ \t]+"([^"]+)")?\)/) ) {
    len = RLENGTH;
    text  = gensub(/^\[([^]]+)\]\(([^"\)]+)([ \t]+"([^"]+)")?\)/, "\\1", "g", line);
    href  = gensub(/^\[([^]]+)\]\(([^"\)]+)([ \t]+"([^"]+)")?\)/, "\\2", "g", line);
    title = gensub(/^\[([^]]+)\]\(([^"\)]+)([ \t]+"([^"]+)")?\)/, "\\4", "g", line);
    if ( title ) {
      return "<a href=\"" HTML(href) "\" title=\"" HTML(title) "\">" inline( text ) "</a>" inline( substr( line, len + 1) );
    } else {
      return "<a href=\"" HTML(href) "\">" inline( text ) "</a>" inline( substr( line, len + 1) );
    }

  # reference style links
  } else if ( match(line, /^\[([^]]+)\] ?\[([^]]*)\]/ ) ) {
    len = RLENGTH;
    text = gensub(/^\[([^\n]+)\] ?\[([^\n]*)\].*/, "\\1", 1, line);
      id = gensub(/^\[([^\n]+)\] ?\[([^\n]*)\].*/, "\\2", 1, line);
    if ( ! id ) id = text;
    if ( rl_href[id] && rl_title[id] ) {
      return "<a href=\"" HTML(rl_href[id]) "\" title=\"" HTML(rl_title[id]) "\">" inline(text) "</a>" inline( substr( line, len + 1) );
    } else if ( rl_href[id] ) {
      return "<a href=\"" HTML(rl_href[id]) "\">" inline(text) "</a>" inline( substr( line, len + 1) );
    } else {
      return "" HTML(substr(line, 1, len)) inline( substr(line, len + 1) );
    }

  # inline images
  } else if ( match(line, /^!\[([^]]+)\]\(([^"\)]+)([ \t]+"([^"]+)")?\)/) ) {
    len = RLENGTH;
    text  = gensub(/^!\[([^]]+)\]\(([^"\)]+)([ \t]+"([^"]+)")?\)/, "\\1", "g", line);
    href  = gensub(/^!\[([^]]+)\]\(([^"\)]+)([ \t]+"([^"]+)")?\)/, "\\2", "g", line);
    title = gensub(/^!\[([^]]+)\]\(([^"\)]+)([ \t]+"([^"]+)")?\)/, "\\4", "g", line);
    if ( title ) {
      return "<img src=\"" HTML(href) "\" alt=\"" HTML(text) "\" title=\"" HTML(title) "\" />" inline( substr( line, len + 1) );
    } else {
      return "<img src=\"" HTML(href) "\" alt=\"" HTML(text) "\" />" inline( substr( line, len + 1) );
    }

  # reference style images
  } else if ( match(line, /^!\[([^]]+)\] ?\[([^]]*)\]/ ) ) {
    len = RLENGTH;
    text = gensub(/^!\[([^\n]+)\] ?\[([^\n]*)\].*/, "\\1", 1, line);
      id = gensub(/^!\[([^\n]+)\] ?\[([^\n]*)\].*/, "\\2", 1, line);
    if ( ! id ) id = text;
    if ( rl_href[id] && rl_title[id] ) {
      return "<img src=\"" HTML(rl_href[id]) "\" alt=\"" HTML(text) "\" title=\"" HTML(rl_title[id]) "\" />" inline( substr( line, len + 1) );
    } else if ( rl_href[id] ) {
      return "<img src=\"" HTML(rl_href[id]) "\" alt=\"" HTML(text) "\" />" inline( substr( line, len + 1) );
    } else {
      return "" HTML(substr(line, 1, len)) inline( substr(line, len + 1) );
    }

  #  ~~strikeout~~ (pandoc)
  } else if ( match(line, /^~~([[:graph:]]|[[:graph:]]([^~]|~[^~])*[[:graph:]])~~/) ) {
    len = RLENGTH;
    return "<del>" inline( substr( line, 3, len - 4 ) ) "</del>" inline( substr( line, len + 1 ) );

  #  ^superscript^ (pandoc)
  } else if ( match(line, /^\^([^[:space:]^]|\\[ ^])+\^/) ) {
    len = RLENGTH;
    return "<sup>" inline( substr( line, 2, len - 2 ) ) "</sup>" inline( substr( line, len + 1 ) );

  #  ~subscript~ (pandoc)
  } else if ( match(line, /^~([^[:space:]~]|\\[ ~])+~/) ) {
    len = RLENGTH;
    return "<sub>" inline( substr( line, 2, len - 2 ) ) "</sub>" inline( substr( line, len + 1 ) );

  # ignore embedded underscores (pandoc, php md)
  } else if ( match(line, "^[[:alnum:]](__|_)") ) {
    return HTML(substr( line, 1, RLENGTH)) inline( substr(line, RLENGTH + 1) );

  #  __strong__$
  } else if ( match(line, "^__(([^_[:space:]]|" ieu ")|([^_[:space:]]|" ieu ")(" nu "|" ieu ")*([^_[:space:]]|" ieu "))__$") ) {
    len = RLENGTH;
    return "<strong>" inline( substr( line, 3, len - 4 ) ) "</strong>" inline( substr( line, len + 1 ) );

  #  __strong__
  } else if ( match(line, "^__(([^_[:space:]]|" ieu ")|([^_[:space:]]|" ieu ")(" nu "|" ieu ")*([^_[:space:]]|" ieu "))__[[:space:][:punct:]]") ) {
    len = RLENGTH;
    return "<strong>" inline( substr( line, 3, len - 5 ) ) "</strong>" inline( substr( line, len) );

  #  **strong**
  } else if ( match(line, "^\\*\\*(([^\\*[:space:]]|" iea ")|([^\\*[:space:]]|" iea ")(" na "|" iea ")*([^\\*[:space:]]|" iea "))\\*\\*") ) {
    len = RLENGTH;
    return "<strong>" inline( substr( line, 3, len - 4 ) ) "</strong>" inline( substr( line, len + 1 ) );

  #  _em_$
  } else if ( match(line, "^_(([^_[:space:]]|" isu ")|([^_[:space:]]|" isu ")(" nu "|" isu ")*([^_[:space:]]|" isu "))_$") ) {
    len = RLENGTH;
    return "<em>" inline( substr( line, 2, len - 2 ) ) "</em>" inline( substr( line, len + 1 ) );

  #  _em_
  } else if ( match(line, "^_(([^_[:space:]]|" isu ")|([^_[:space:]]|" isu ")(" nu "|" isu ")*([^_[:space:]]|" isu "))_[[:space:][:punct:]]") ) {
    len = RLENGTH;
    return "<em>" inline( substr( line, 2, len - 3 ) ) "</em>" inline( substr( line, len ) );

  #  *em*
  } else if ( match(line, "^\\*(([^\\*[:space:]]|" isa ")|([^\\*[:space:]]|" isa ")(" na "|" isa ")*([^\\*[:space:]]|" isa "))\\*") ) {
    len = RLENGTH;
    return "<em>" inline( substr( line, 2, len - 2 ) ) "</em>" inline( substr( line, len + 1 ) );

  # Macros
  } else if ( AllowMacros && match( line, /^<<([^>]|>[^>])+>>/) ) {
    len = RLENGTH;
    return macro( substr( line, 3, len - 4 ) ) inline(substr(line, len + 1));

  # Verbatim inline HTML
  } else if ( AllowHTML && match( line, /^(<!--([^-]|-[^-]|--[^>])*-->|<\?([^\?]|\?[^>])*\?>|<![A-Z][^>]*>|<!\[CDATA\[([^\]]|\][^\]]|\]\][^>])*\]\]>|<\/[A-Za-z][A-Za-z0-9-]*[[:space:]]*>|<[A-Za-z][A-Za-z0-9-]*([[:space:]]+[A-Za-z_:][A-Za-z0-9_\.:-]*([[:space:]]*=[[:space:]]*([[:space:]"'=<>`]+|"[^"]*"|'[^']*'))?)*[[:space:]]*\/?>)/) ) {
    len = RLENGTH;
    return substr( line, 1, len) inline(substr(line, len + 1));

  # Literal HTML entities
  } else if ( match( line, /^&([a-zA-Z]{2,32}|#[0-9]{1,7}|#[xX][0-9a-fA-F]{1,6});/) ) {
    len = RLENGTH;
    return substr( line, 1, len ) inline(substr(line, len + 1));

  # Escape lone HTML character
  } else if ( match( line, /^[&<>"']/) ) {
    return HTML(substr(line, 1, 1)) inline(substr(line, 2));

  #  continue walk over string
  } else {
    return substr(line, 1, 1) inline( substr(line, 2) );
  }
}

function _block( block, LOCAL, st, len, hlvl, htxt, guard, code, indent, attrib ) {
  gsub( /^\n+|\n+$/, "", block );

  if ( block == "" ) {
    return "";

  # HTML #2 #3 #4 $5
  } else if ( AllowHTML && match( block, /(^|\n) ? ? ?(<!--([^-]|-[^-]|--[^>])*(-->|$)|<\?([^\?]|\?[^>])*(\?>|$)|<![A-Z][^>]*(>|$)|<!\[CDATA\[([^\]]|\][^\]]|\]\][^>])*(\]\]>|$))/) ) {
    len = RLENGTH; st = RSTART;
    return _block(substr(block, 1, st - 1)) substr(block, st, len) _block(substr(block, st + len));

  # HTML #6
  } else if ( AllowHTML && match( tolower(block), /(^|\n) ? ? ?<\/?(address|article|aside|base|basefont|blockquote|body|caption|center|col|colgroup|dd|details|dialog|dir|div|dl|dt|fieldset|figcaption|figure|footer|form|frame|frameset|h[123456]|head|header|hr|html|iframe|legend|li|link|main|menu|menuitem|nav|noframes|ol|optgroup|option|p|param|section|source|summary|table|tbody|td|tfoot|th|thead|title|tr|track|ul)([[:space:]\n>]|\/>)([^\n]|\n[ \t]*[^\n])*(\n[[:space:]]*\n|$)/) ) {
    len = RLENGTH; st = RSTART;
    return _block(substr(block, 1, st - 1)) substr(block, st, len) _block(substr(block, st + len));

  # HTML #1
  } else if ( AllowHTML && match( tolower(block), /(^|\n) ? ? ?<(script|pre|style)([[:space:]\n>]).*(<\/script>|<\/pre>|<\/style>|$)/) ) {
    len = RLENGTH; st = RSTART;
    match( tolower(substr(block, st, len)), /(<\/script>|<\/pre>|<\/style>)/);
    len = RSTART + RLENGTH;
    return _block(substr(block, 1, st - 1)) substr(block, st, len) _block(substr(block, st + len));

  # HTML #7
  } else if ( AllowHTML && match( block, /^ ? ? ?(<\/[A-Za-z][A-Za-z0-9-]*[[:space:]]*>|<[A-Za-z][A-Za-z0-9-]*([[:space:]]+[A-Za-z_:][A-Za-z0-9_\.:-]*([[:space:]]*=[[:space:]]*([[:space:]"'=<>`]+|"[^"]*"|'[^']*'))?)*[[:space:]]*\/?>)([[:space:]]*\n)([^\n]|\n[ \t]*[^\n])*(\n[[:space:]]*\n|$)/) ) {
    len = RLENGTH; st = RSTART;
    return substr(block, st, len) _block(substr(block, st + len));
 
  # Blockquote (leading >)
  } else if ( match( block, /^> /) ) {
    match( block, /(^|\n)[[:space:]]*(\n|$)/ ) || match(block, /$/);
    len = RLENGTH; st = RSTART;
    return "<blockquote>\n" _block( gensub( /(^|\n)> /, "\n", "g", substr(block, 1, st - 1) ) ) "</blockquote>\n\n" \
           _block( substr(block, st + len) );

  # Line Blocks (pandoc)
  } else if ( match(block, /^\| [^\n]*(\n|$)(\| [^\n]*(\n|$)|[ \t]+[^\n[:space:]][^\n]*(\n|$))*/) ) {
    len = RLENGTH; st = RSTART;
    code = substr(block, 1, len);
    gsub(/\n[[:space:]]+/, " ", code);
    gsub(/\n\| /, "\n", code);
    gsub(/^\| |\n$/, "", code);
    return "<div class=\"line-block\">" gensub(/\n/, "<br />\n", "g", inline( code )) "</div>\n" \
           _block( substr( block, len + 1) );

  # Indented Code Block
  } else if ( match(block, /^(    |\t)[^\n]+(\n|$)((    |\t)[^\n]+(\n|$)|[ \t]*(\n|$))*/) ) {
    len = RLENGTH; st = RSTART;
    code = substr(block, 1, len);
    gsub(/(^|\n)(    |\t)/, "\n", code);
    gsub(/^\n|\n+$/, "", code);
    return "<pre><code>" HTML( code ) "</code></pre>\n" \
           _block( substr( block, len + 1 ) );

  # Fenced Divs (pandoc, custom)
  } else if ( match( block, /^(:::+)/ ) ) {
    guard = substr( block, 1, RLENGTH );
    code = gensub(/^[^\n]+\n/, "", 1, block);
    attrib = gensub(/^:::+[ \t]*\{?[ \t]*([^\}\n]*)\}?[ \t]*\n.*$/, "\\1", 1, block);
    gsub(/[^a-zA-Z0-9_-]+/, " ", attrib);
    gsub(/(^ | $)/, "", attrib);
    if ( match(code, "(^|\n)" guard "+(\n|$)" ) ) {
      len = RLENGTH; st = RSTART;
      return "<div class=\"" attrib "\">" _block( substr(code, 1, st - 1) ) "</div>\n" \
             _block( substr( code, st + len ) );
    } else {
      match( block, /(^|\n)[[:space:]]*(\n|$)/ ) || match( block, /$/ );
      len = RLENGTH; st = RSTART;
      return "<p>" inline( substr(block, 1, st - 1) ) "</p>\n" \
             _block( substr(block, st + len) );
    }

  # Fenced Code Block (pandoc)
  } else if ( match( block, /^(~~~+|```+)/ ) ) {
    guard = substr( block, 1, RLENGTH );
    code = gensub(/^[^\n]+\n/, "", 1, block);
    attrib = gensub(/^:::+[ \t]*\{?[ \t]*([^\}\n]*)\}?[ \t]*\n.*$/, "\\1", 1, block);
    gsub(/[^a-zA-Z0-9_-]+/, " ", attrib);
    gsub(/(^ | $)/, "", attrib);
    if ( match(code, "(^|\n)" guard "+(\n|$)" ) ) {
      len = RLENGTH; st = RSTART;
      return "<pre><code class=\"" attrib "\">" HTML( substr(code, 1, st - 1) ) "</code></pre>\n" \
             _block( substr( code, st + len ) );
    } else {
      match( block, /(^|\n)[[:space:]]*(\n|$)/ ) || match( block, /$/ );
      len = RLENGTH; st = RSTART;
      return "<p>" inline( substr(block, 1, st - 1) ) "</p>\n" \
             _block( substr(block, st + len) );
    }

  # Unordered list
  } else if ( match( block, "^ ? ? ?[-+*][ \t]+[^\n]+(\n|$)" \
                            "(([ \t]*\n)* ? ? ?[-+*][ \t]+[^\n]+(\n|$)" \
                            "|([ \t]*\n)*( ? ? ?\t|  +)[^\n]+(\n|$)" \
                            "|[^\n]+(\n|$))*" ) ) {
  list = substr( block, 1, RLENGTH);
  block = substr( block, RLENGTH + 1);
  indent = length( gensub(/[-+*][ \t]+[^\n]+.*$/, "", 1, list) );

  gsub("(^|\n) {0," indent "}", "\n", list);
  return "\n<ul>\n" _list( substr(list, 2) ) "</ul>\n" _block( block );

  # Ordered list
  } else if ( match( block, "^ ? ? ?([0-9]+|#)\\.[ \t]+[^\n]+(\n|$)" \
                            "(([ \t]*\n)* ? ? ?([0-9]+|#)\\.[ \t]+[^\n]+(\n|$)" \
                            "|([ \t]*\n)*( ? ? ?\t|  +)[^\n]+(\n|$)" \
                            "|[^\n]+(\n|$))*" ) ) {
  list = substr( block, 1, RLENGTH);
  block = substr( block, RLENGTH + 1);
  indent = length( gensub(/([0-9]+|#)\.[ \t]+[^\n]+.*$/, "", 1, list) );

  gsub("(^|\n) {0," indent "}", "\n", list);
  return "\n<ol>\n" _list( substr(list, 2) ) "</ol>\n" _block( block );

  # First Order Heading
  } else if ( match( block, /^[^\n]+\n===+(\n|$)/ ) ) {
    len = RLENGTH;
    HL[1]++; HL[2] = 0; HL[3] = 0; HL[4] = 0; HL[5] = 0; HL[6] = 0;
    return "<h1 id=\"" HL[1] " - " HTML(gensub( /\n.*$/, "", "g", block )) "\">" inline( gensub( /\n.*$/, "", "g", block ) ) "</h1>\n\n" \
           _block( substr( block, len + 1 ) );

  # Second Order Heading
  } else if ( match( block, /^[^\n]+\n---+(\n|$)/ ) ) {
    len = RLENGTH;
    HL[2]++; HL[3] = 0; HL[4] = 0; HL[5] = 0; HL[6] = 0;
    return "<h2 id=\"" HL[1] "." HL[2] " - " HTML(gensub( /\n.*$/, "", "g", block )) "\">" inline( gensub( /\n.*$/, "", "g", block ) ) "</h2>\n\n" \
           _block( substr( block, len + 1) );

  # Nth Order Heading
  } else if ( match( block, /^#{1,6}[ \t]*[^\n]+([ \t]*#*)(\n|$)/ ) ) {
    len = RLENGTH;
    hlvl = length( gensub( /^(#{1,6}).*$/, "\\1", "g", block ) );
    htxt = gensub(/^#{1,6}[ \t]*(([^ \t\n]+|[ \t]+[^ \t\n#]|[ \t]+#+[^\n#])+)([ \t]*#*)(\n.*)?$/, "\\1", 1, block);
    HL[hlvl]++; for ( n = hlvl + 1; n < 7; n++) { HL[n] = 0;}
    hid = HL[1]; for ( n = 2; n <= hlvl; n++) { hid = hid "." HL[n] ; }
    return "<h" hlvl " id=\"" hid " - " HTML(htxt) "\">" inline( htxt ) "</h" hlvl ">\n\n" \
           _block( substr( block, len + 1) );

  # Horizontal rule
  } else if ( match( block, /(^|\n) ? ? ?((\* *){3,}|(- *){3,}|(_ *){3,})($|\n)/) ) {
    len = RLENGTH; st = RSTART;
    return _block(substr(block, 1, st - 1)) "<hr />\n" _block(substr(block, st + len));

  # Plain paragraph
  } else {
    match( block, /(^|\n)[[:space:]]*(\n|$)/ ) || match( block, /$/ );
    len = RLENGTH; st = RSTART;
    return "<p>" inline( substr(block, 1, st - 1) ) "</p>\n" \
           _block( substr(block, st + len) );
  }
}

function _list( block, last, LOCAL, p) {
  if ( ! length(block) ) return "";
  gsub(/^([-+*]|[0-9]+\.|#\.)(  ? ? ?|\t)/, "", block)

  # slice next list item from input
  if ( match( block, /\n([-+*]|[0-9]+\.|#\.)[ \t]+[^\n]+/) ) {
    p = substr( block, 1, RSTART);
    block = substr( block, RSTART + 1);
  } else {
    p = block; block = "";
  }
  sub( /\n +([-+*]|[0-9]+\.|#\.)/, "\n&", p );

  # if this should be a paragraph item
  # either previous item (last) or current item (p) contains blank lines
  if (match(last, /\n[[:space:]]*\n/) || match(p, /\n[[:space:]]*\n/) ) {
    last = p; p = _block(p);
  } else {
    last = p; p = _block(p);
    sub( /^<p>/, "", p );
    sub( /<\/p>\n/, "", p );
  }
  sub( /\n$/, "", p );

  # Task List (pandoc)
       if ( p ~ /^\[ \].*/ )       { p = "<input type=checkbox disabled />" substr(p, 4); }
  else if ( p ~ /^\[[xX]\].*/ )    { p = "<input type=checkbox disabled checked />" substr(p, 4); }
  else if ( p ~ /^<p>\[ \].*/ )    { p = "<p><input type=checkbox disabled />" substr(p, 7); }
  else if ( p ~ /^<p>\[[xX]\].*/ ) { p = "<p><input type=checkbox disabled checked />" substr(p, 7); }
  return "<li>" p "</li>\n" _list( block, last );
}

BEGIN {
  # Global Vars
  file = ""; rl_href[""] = ""; rl_title[""] = "";
  if (ENVIRON["MD_HTML"] == "true") { AllowHTML = "true"; }
  HL[1] = 0; HL[2] = 0; HL[3] = 0; HL[4] = 0; HL[5] = 0; HL[6] = 0;

  # Buffering of full file ist necessary, e.g. to find reference links
  while (getline) { file = file $0 "\n"; }
  # Clean up MS-DOS line breaks
  gsub(/\r\n/, "\n", file);

  # Fill array of reference links
  f = file; rl_id;
  re_reflink = "(^|\n) ? ? ?\\[([^]\n]+)\\]: ([^ \t\n]+)(\n?[ \t]+(\"([^\"]+)\"|'([^']+)'|\\(([^)]+)\\)))?(\n|$)";
  # /(^|\n) ? ? ?\[([^]\n]+)\]: ([^ \t\n]+)(\n?[ \t]+("([^"]+)"|'([^']+)'|\(([^)]+)\)))?(\n|$)/
  while ( match(f, re_reflink ) ) {
    rl_id           = gensub( re_reflink, "\\2", 1, substr(f, RSTART, RLENGTH) );
    rl_href[rl_id]  = gensub( re_reflink, "\\3", 1, substr(f, RSTART, RLENGTH) );
    rl_title[rl_id] = gensub( re_reflink, "\\5", 1, substr(f, RSTART, RLENGTH) );
    f = substr(f, RSTART + RLENGTH);
    rl_title[rl_id] = substr( rl_title[rl_id], 2, length(rl_title[rl_id]) - 2 );
    if ( rl_href[rl_id] ~ /<.*>/ ) rl_href[rl_id] = substr( rl_href[rl_id], 2, length(rl_href[rl_id]) - 2 );
  }
  # Clear reflinks from File
  while( gsub(re_reflink, "\n", file ) );
  # for (n in rl_href) { debug(n " | " rl_href[n] " | " rl_title[n] ); }

  # Run Block Processing -> The Actual Markdown!
  printf "%s", _block( file );
}
